# This is a blank file used to allow import from subdirectories.
# References:
# https://stackoverflow.com/questions/1260792/import-a-file-from-a-subdirectory
# https://packaging.python.org/tutorials/packaging-projects/#a-simple-project

# Add import pdbp to an __init__.py of your project, which
# will automatically make Pdb+ the default debugger at breakpoints
# import pdbp
